/**
  * ---------------------------------------------------------------------------
  * ~k radix implementation // amethyst nanokernel project
  * ---------------------------------------------------------------------------
  * author: DG.Ward (elderK)
  * contact: 
  *   kyretzn@gmail.com
  *   comeherenightmare@hotmail.com
  *   warddg1@student.op.ac.nz
  *
  * date: May 15th, 2008.
  * version: 0.0.1
  * ---------------------------------------------------------------------------
  * All source herein is licensed under the Creative Commons : Attribution,
  * No Derivative Works license. For more information, please go to this
  * website: http://creativecommons.org/licenses/by-nd/3.0/
  * ---------------------------------------------------------------------------
  * This software is powered by NINE INCH NAILS :^)
  * ---------------------------------------------------------------------------
  * 
  * ---------------------------------------------------------------------------
  * INSERTION EXAMPLE
  * ---------------------------------------------------------------------------
  *   
  *     // kradix does not have any special "tree" nodes. Instead,
  *     // a regular radix pointer is used. It must be initialized
  *     // to NULL. If it is not, insert_node(...) will not recognize it
  *     // as a new tree and try to find an insertion point somewhere in
  *     // memory ... god knows where ...
  *     radix_node_t *a_tree = NULL;
  *
  *     // Insert a node, bound to no data, identified by the string "test".
  *     // Note &a_tree - we pass the root pointers ADDRESS. This is so
  *     // insert_node(...) can itself, directly update what root references.
  *     if(insert_node(&a_tree, "test", NULL))
  *     {
  *         // insertion succeeded.
  *     }
  *
  **/
#include "radix.h"

/**
  * place_node(external, internal, iterator)
  *
  *     this call is what really does the work of insertion.
  *
  *     it is responsible for:
  *         - isolating the difference between the string being
  *           inserted and that of the insertion point.
  *
  *         - recognizing and correctly ordering prefixes.
  *
  *     parameters:
  *         external = reference to the external node being
  *                    inserted.
  *         internal = reference to the internal node which will
  *                    detail the differences between external and
  *                    iterator.
  *         iterator = reference to the iterator node, the node
  *                    that currently inhabits the insertion point.
  *
  *     returns:
  *         a reference to the internal node.
  *
  **/
static
radix_node_t
*place_node
(
    radix_node_t *external,
    radix_node_t *internal,
    radix_node_t *iterator
)
{
    /* these two pointers will be used as a shortcut later */
    char *exstr, *itstr;
    /* later, this will store the difference between bytes */
    uint8_t diff_cache;

    /*
        compare_len will point to the smaller strlen.

        diff_pos will be used to determine which node stores the longer
        string and thus, which is the potential prefix; external or
        iterator.
    */
    uint32_t *compare_len, diff_pos =
        extern(iterator).ne_identifier_len >
            extern(external).ne_identifier_len;

    if(diff_pos)
        compare_len = &extern(iterator).ne_identifier_len;
    else
        compare_len = &extern(external).ne_identifier_len;

    /* link the pointers, so we dont have to type so much later ... */
    exstr = extern(external).ne_identifier;
    itstr = extern(iterator).ne_identifier;

    /*
        compare the two identifiers; compare only up to the smallest
        strings length. Break to NO_MATCH when a difference is detected.
        If no difference is detected, fall through.
    */
    for(diff_pos = 0; diff_pos < *compare_len; diff_pos++)
        if(exstr[diff_pos] ^ itstr[diff_pos])
            goto no_match;

    /* match: */
        /*
            it has been decided that one of the strings, is a prefix.
            thats all well and jolly, but we need to know WHICH is the
            prefix. Iterator or External?
        */
        diff_cache = (compare_len == &extern(external).ne_identifier_len);

        /* mark the internal node, signalling that it details a prefix/es */
        internal->n_flag = NT_INTERN_MARKED;

    place:
        /*
            setup the internal nodes references...

            diff_cache stores the "direction" of the external node.
            Left = 0, Right = 1.

            As mentioned in the radix.h header, ni_diff_pos's interpretation
            differs depending on whether the internal node is marked or not.
            *compare_len points to the prefix length (if prefix case) and the
            bit index of difference (if not prefix).
        */
        intern(internal).ni_diff_pos = *compare_len;
        intern(internal).ni_link[diff_cache] = external;
        intern(internal).ni_link[!diff_cache] = iterator;

        return internal;

    no_match:
        /* cache the differing byte */
        diff_cache = (uint8_t)(exstr[diff_pos] ^ itstr[diff_pos]);
        
        /* isolate the bit position of the difference */
        for(diff_pos <<= 3; !(diff_cache & 1); diff_cache >>= 1, diff_pos++);

        /* determine which side of the internal node external will be placed */
        diff_cache = (exstr[diff_pos >> 0x3] & (1 << (diff_pos & 0x7))) > 0;

        /*
            ensure the internal node is UNMARKED and that compare_len points
            to the correct data; in this case, the bit index of the difference.
        */
        internal->n_flag = NT_INTERN_UNMARKED;
        compare_len = &diff_pos;
        goto place;
}

/**
  * insert_node(&tree, identifier, data)
  *
  *     this function is responsible for everything insertion!
  *     meaning, it creates both an internal and external node,
  *     marks the internal node if necessary and places the
  *     external node itself.
  *
  *     parameters:
  *        &tree = root pointer pointer.
  *        identifier = ascii string identifier.
  *        data = pointer to data to be bound by identifier.
  *
  *     return values:
  *        NULL, on FAILURE.
  *        A reference to the inserted node, on SUCCESS.
  *
  *     see also:
  *        an example of this call can be found in insert_node.c
  *
  *     NOTE: A copy of the identifier string is _NOT_ created.
  *     The node's ne_identifier reference will be bound to
  *     precisely the IDENTIFIER parameter passed.
  *
  *     The node's ne_data field acts the same with the DATA
  *     paramater. The responsibility to dynamically allocate
  *     persistent copies is the User's burden.
  *
  **/
radix_node_t
*insert_node(radix_node_t **tree, char *identifier, void *data)
{
    /* 
        every insertion, bar the "dead root" case, requires an internal
        node - to detail differences between identifiers.
    */
    radix_node_t *internal = create_node();
    if (internal)
    {
        /* create a node to represent the new key/data pair */
        radix_node_t *external = create_node(), *iterator = *tree;
            
        if (!external)
        {
            release_node(internal);
            return NULL;
        }

        /* set the external nodes attributes */
        extern(external).ne_identifier = identifier;
        extern(external).ne_identifier_len = strlen(identifier);
    
        /* Is the tree alive? (iterator = *tree) */
        if (iterator)
        {
            /* Yes, create a pointer to keep track of the iterator's parent */
            radix_node_t **parent = tree;

            /* Traverse the tree, searching for an insertion point. */
            for(; iterator->n_flag != NT_EXTERN; iterator = *parent)
            {
                /* Marked or Unmarked Internal node? */
                if((iterator->n_flag & NT_MASK) == NT_INTERN_UNMARKED)
                {
                    /*
                        check the bit state specified by ni_diff_pos, in
                        identifier. If it is set, we route right. If clear,
                        we route left.
                    */
                    internal->n_flag =
                        identifier[intern(iterator).ni_diff_pos >> 0x3]
                            & (1 << (intern(iterator).ni_diff_pos & 0x7));
                }
                else
                {
                    /*
                        determine which subtree of the marked node that
                        identifier best belongs in. If it is smaller than
                        the length specified in ni_diff_pos, it can live
                        in the prefix subtree. If not, it can live in the
                        big-kids subtree ... 
                    */
                    internal->n_flag =
                        intern(iterator).ni_diff_pos <=
                            extern(external).ne_identifier_len;
                }

                /* keep the parent pointer up to date */
                parent = &intern(iterator).ni_link[internal->n_flag > 0];
            }

            /* finalize the insertion; link in the internal node */
            *parent = place_node(external, internal, iterator);
        }
        else
        {
            /*
                No, the root pointer identifies a dead tree.
                We no longer need the internal node, release it.
            */ 
            release_node(internal);
            *tree = external;
        }

        /* Bind the data to the external node */
        extern(external).ne_data = data;
        return external;
    }
    return NULL;
}
