/**
  * ---------------------------------------------------------------------------
  * ~k radix implementation // amethyst nanokernel project
  * ---------------------------------------------------------------------------
  * author: DG.Ward (elderK)
  * contact: 
  *   kyretzn@gmail.com
  *   comeherenightmare@hotmail.com
  *   warddg1@student.op.ac.nz
  *
  * date: May 15th, 2008.
  * version: 0.0.1
  * ---------------------------------------------------------------------------
  * All source herein is licensed under the Creative Commons : Attribution,
  * No Derivative Works license. For more information, please go to this
  * website: http://creativecommons.org/licenses/by-nd/3.0/
  * ---------------------------------------------------------------------------
  * This software is powered by NINE INCH NAILS :^)
  * ---------------------------------------------------------------------------
  *
  * ---------------------------------------------------------------------------
  * LOCATE EXAMPLE
  * ---------------------------------------------------------------------------
  *   
  * // in order to lookup the desired node, we must create a trace
  * // structure. This structure stores all the information needed to
  * // remove the node, should we want to. It also stores a reference
  * // to our desired node, of course :).
  * k_radix_trace_t lookup_result;
  *
  * // the pointer, root, in this example references a valid tree.
  * lookup_result = locate_node(&root, "test_ident");
  *
  * // k_radix_trace_t.t_frame[0] signals whether the lookup was
  * // successful or not.
  * if(lookup_result.t_frame[0])
  * {
  *      // lookup was successful ...
  * }
  *
  **/
#include "radix.h"

/**
  * locate_node(&tree, identifier)
  *
  *     this function is used to retrieve a reference to
  *     a node bound to a specific identifier.
  *
  *     parameters:
  *         &tree = root pointer pointer.
  *         identifier = ascii string identifer of desired node.
  *
  *     return:
  *         in all cases, this call will return a radix_trace_t
  *         structure.
  *
  *         On failure, t_frame[0] of the structure will be NULL.
  *         On success, t_frame[0] references the desired node.
  *
  *     see also:
  *         an example on the use of this function can be found in
  *         locate_node.c
  *
  **/
radix_trace_t
locate_node(radix_node_t **tree, char *identifier)
{
    radix_node_t *iterator = *tree;

    /* create, initialize trace_frame. */
    radix_trace_t trace_frame = {{NULL}};

    /* is the tree alive / iterator valid? */
    if(iterator)
    {
        /* cache the length of identifier */
        uint32_t ident_len = strlen(identifier);

        /* preset t_frame[0]; first node visited */
        trace_frame.t_frame[0] = iterator;

        /*
            Traverse the tree, routing based on the input identifier.
            Stop once we arrive at an external node to compare with.
        */
        while(iterator->n_flag != NT_EXTERN)
        {
            /* Marked or Unmarked? */
            if((iterator->n_flag & NT_MASK) == NT_INTERN_MARKED)
            {
                /*
                    Marked. Traverse to the most suitable subtree.
                    ie: String smaller than marked node demands,
                    route to the prefix subtree. String greater,
                    route to the nonprefix subtree.
                */
                tree = &intern(iterator).ni_link[ident_len <=
                    intern(iterator).ni_diff_pos];
            }
            else
            {
                /*
                    Determine the state of the bit specified by
                    ni_diff_pos, within the identifier string. If set,
                    route right. If clear, route left.
                */
                tree = &intern(iterator).ni_link[
                    (identifier[intern(iterator).ni_diff_pos >> 3] &
                        (1 << (intern(iterator).ni_diff_pos & 7))) > 0];
            }

            /* Update the trace stack */
            trace_frame.t_frame[2] = trace_frame.t_frame[1];
            trace_frame.t_frame[1] = trace_frame.t_frame[0];
            trace_frame.t_frame[0] = iterator = *tree;
        }

        /* Make sure we've located the desired node */
        if(strncmp(identifier, extern(iterator).ne_identifier, ident_len))
            memset(&trace_frame, 0x00, sizeof(radix_trace_t));
    }

    return trace_frame;
}
