/**
  * ---------------------------------------------------------------------------
  * ~k radix implementation // amethyst nanokernel project
  * ---------------------------------------------------------------------------
  * author: DG.Ward (elderK)
  * contact: 
  *   kyretzn@gmail.com
  *   comeherenightmare@hotmail.com
  *   warddg1@student.op.ac.nz
  *
  * date: May 15th, 2008.
  * version: 0.0.1
  * ---------------------------------------------------------------------------
  * All source herein is licensed under the Creative Commons : Attribution,
  * No Derivative Works license. For more information, please go to this
  * website: http://creativecommons.org/licenses/by-nd/3.0/
  * ---------------------------------------------------------------------------
  * This software is powered by NINE INCH NAILS :^)
  * ---------------------------------------------------------------------------
  *
  **/
#ifndef _H_K_RADIX
#define _H_K_RADIX 1

#define __HOSTED__ 1

#ifdef __HOSTED__
    #include <stdlib.h>
#endif

#include <string.h>
#include <stdint.h>

/**
  * radix_node_t
  *
  *     here defined is the general purpose node structure
  *     for the K Radix implementation. The single structure
  *     acts as both external and internal node, depending on
  *     the value held within n_flag.
  *
  *     the implementation only uses the two least significant
  *     of n_flag - the remaining thirty bits are available for
  *     application-specific use.
  *
  *     prefix-map:
  *     n ni ne nt ---> node type
  *     | |  `--------> node external
  *     | `-----------> node internal
  *     `-------------> node
  *
  *     individual members of the composite are explained below.
  *
  **/
typedef
struct radix_node_t
{
    /* n_flag is used to identify the node's current class. */
    uint32_t n_flag;

    /*
        n_type_sel is used to change the interpretation of the
        core node structures at will. Structures within this union,
        obviously, must be the same size.
    */
    union
    {
        /* n_extern = leaf node, stores data */
        struct
        {
            /* data that is related to the key */
            void *ne_data;

            /* identifier, ascii string */
            char *ne_identifier;

            /* length of identifier ... */
            uint32_t ne_identifier_len;
        }
        n_extern;

        /* n_intern = internal node, stores route/lookup information */
        struct
        {
            /* children links */
            struct radix_node_t *ni_link[2];

            /* 
               how this member is interpreted depends whether the
               internal node is MARKED or UNMARKED. when the node
               is not marked, ni_diff_pos specifies the bit-index to
               branch on.

               When the node is marked, it specifies the prefix length
               in bytes.
            */
            uint32_t ni_diff_pos;
        }
        n_intern;
    }
    n_type_sel;
}
__attribute__((packed))
radix_node_t;

/* friendly neighbourhood macros to make life easier ... */
#define intern(ni)  ni->n_type_sel.n_intern
#define extern(ne)  ne->n_type_sel.n_extern

/* n_flag type definitions. */
enum
{
    NT_EXTERN           = 0x0000,
    NT_INTERN_UNMARKED  = 0x0001,
    NT_INTERN_MARKED    = 0x0003,
    NT_MASK             = 0x0003
}
radix_type_t;

/**
  * radix_trace_t
  *
  *     the radix_trace_t structure is used during both
  *     the removal and lookup of nodes within the radix tree.
  * 
  *     its sole point is to provide both the target node and
  *     that node's immediate relations (parent, grandparent...)
  *
  *     Trace frame indices are allocated as follow:
  *     0 - Target of lookup operation (target node).
  *     1 - Target's parent
  *     2 - Target's grandparent
  *
  **/
typedef
struct radix_trace_t
{
    radix_node_t *t_frame[3];
}
__attribute__((packed))
radix_trace_t;

/**
  * create_node()
  *
  *     create_node() serves only to allocate and initialize
  *     a node for use.
  *
  *     the call takes zero parameters and will return NULL
  *     on error. On success, it returns a reference to the
  *     newly created / initialized node.
  *
  **/
extern radix_node_t *create_node(void);

/**
  * release_node(node)
  *
  *     the alterego of create_node(). This call exists only to
  *     release the resources tied to the node passed.
  *
  *     NOTE: This will __NOT__ deallocate any data referenced
  *     by the node's ne_data or ne_identifier members!
  *
  *     The responsibility to deallocate bound data rests on the
  *     shoulders of the User.
  *
  **/
extern void release_node(radix_node_t *);

/**
  * insert_node(&tree, identifier, data)
  *
  *     this function is responsible for everything insertion!
  *     meaning, it creates both an internal and external node,
  *     marks the internal node if necessary and places the
  *     external node itself.
  *
  *     parameters:
  *        &tree = root pointer pointer.
  *        identifier = ascii string identifier.
  *        data = pointer to data to be bound by identifier.
  *
  *     return values:
  *        NULL, on FAILURE.
  *        A reference to the inserted node, on SUCCESS.
  *
  *     see also:
  *        an example of this call can be found in insert_node.c
  *
  *     NOTE: A copy of the identifier string is _NOT_ created.
  *     The node's ne_identifier reference will be bound to
  *     precisely the IDENTIFIER parameter passed.
  *
  *     The node's ne_data field acts the same with the DATA
  *     paramater. The responsibility to dynamically allocate
  *     persistent copies is the User's burden.
  *
  **/
extern radix_node_t *insert_node(radix_node_t **, char *, void *);

/**
  * locate_node(&tree, identifier)
  *
  *     this function is used to retrieve a reference to
  *     a node bound to a specific identifier.
  *
  *     parameters:
  *         &tree = root pointer pointer.
  *         identifier = ascii string identifer of desired node.
  *
  *     return:
  *         in all cases, this call will return a radix_trace_t
  *         structure.
  *
  *         On failure, t_frame[0] of the structure will be NULL.
  *         On success, t_frame[0] references the desired node.
  *
  *     see also:
  *         an example on the use of this function can be found in
  *         locate_node.c
  *
  **/
extern radix_trace_t locate_node(radix_node_t **, char *);

/**
  * remove_node(&tree, trace)
  *
  *     this function is responsible for correctly removing the node
  *     identified by the trace radix_trace_t structure.
  *
  *     for this call to function correctly, the trace structure
  *     passed, must not have been modified after locating the node
  *     to which it relates.
  *
  *     the sequence for removing a node is as follows:
  *     i) Locate the node, resulting in trace data.
  *     ii) Remove node, passing in gathered trace.
  *
  *     parameters:
  *         &tree = root pointer pointer.
  *         trace = pointer to trace detailing the victim node.
  *
  *     return:
  *         a reference to the node removed, no longer in the tree.
  *
  *     see also:
  *         examples on how to use this function properly are given
  *         in remove_node.c
  *
  **/
extern radix_node_t *remove_node(radix_node_t **, radix_trace_t *);

#endif
