/**
  * ---------------------------------------------------------------------------
  * ~k radix implementation // amethyst nanokernel project
  * ---------------------------------------------------------------------------
  * author: DG.Ward (elderK)
  * contact: 
  *   kyretzn@gmail.com
  *   comeherenightmare@hotmail.com
  *   warddg1@student.op.ac.nz
  *
  * date: May 15th, 2008.
  * version: 0.0.1
  * ---------------------------------------------------------------------------
  * All source herein is licensed under the Creative Commons : Attribution,
  * No Derivative Works license. For more information, please go to this
  * website: http://creativecommons.org/licenses/by-nd/3.0/
  * ---------------------------------------------------------------------------
  * This software is powered by NINE INCH NAILS :^)
  * ---------------------------------------------------------------------------
  *
  * ---------------------------------------------------------------------------
  * REMOVAL EXAMPLE
  * ---------------------------------------------------------------------------
  *   
  * // The first step to removing a node, is gathering the trace
  * // information. This information details the node's parents and is
  * // required to correctly relink the tree.
  * k_radix_trace_t trace_info;
  *
  * // Of course, to get this information, we need to locate the node.
  * // In this example, the root pointer references a valid tree.
  * trace_info = locate_node(&root, "test_identifier");
  *
  * // Check to see if the lookup was successful.
  * if(trace_info.t_frame[0] == NULL)
  * {
  *     // Couldnt find "test_identifier" ...
  * }
  *
  * // If we successfully located the node, we can now remove it!
  * // Don't worry, the removal call does not have to relocate the node.
  * radix_node_t *victim = remove_node(&root, &trace_info);
  *
  * // victim is now the only entity to reference node "test_identifier".
  * // So, lets set it free ...
  * release_node(victim);
  *
  **/
#include "radix.h"

/**
  * remove_node(&tree, trace)
  *
  *     this function is responsible for correctly removing the node
  *     identified by the trace radix_trace_t structure.
  *
  *     for this call to function correctly, the trace structure
  *     passed, must not have been modified after locating the node
  *     to which it relates.
  *
  *     the sequence for removing a node is as follows:
  *     i) Locate the node, resulting in trace data.
  *     ii) Remove node, passing in gathered trace.
  *
  *     parameters:
  *         &tree = root pointer pointer.
  *         trace = pointer to trace detailing the victim node.
  *
  *     return:
  *         a reference to the node removed, no longer in the tree.
  *
  *     see also:
  *         examples on how to use this function properly are given
  *         in remove_node.c
  *
  **/
radix_node_t
*remove_node(radix_node_t **tree, radix_trace_t *trace)
{
    /* Is this node the root? */
    if(trace->t_frame[1])
    {
        /*
            No, its not. Find out which child of the internal node we
            are removing.
        */
        uint8_t extern_branch =
            trace->t_frame[0] == intern(trace->t_frame[1]).ni_link[1];

        /* Is the internal node root? */
        if(trace->t_frame[2])
        {
            /* No, find out which child the internal node is. */
            uint8_t intern_branch =
                trace->t_frame[1] == intern(trace->t_frame[2]).ni_link[1];

            /*
                Tag the Grandparent's pointer to Internal node, after
                removal it will reference the victim node's sibling.
            */
            tree = &intern(trace->t_frame[2]).ni_link[intern_branch];
        }

        /*
            t_frame[2] stores the address, that will be patched into
            the parent's link to the victim node.
        */
        trace->t_frame[2] =
            intern(trace->t_frame[1]).ni_link[!extern_branch];

        /* Set the Internal node free ... */    
        release_node(trace->t_frame[1]);
    }

    /*
        Update the *tree's reference. If Internal was root, the root pointer
        passed into the function will be updated. Otherwise, the grandparent
        will be updated.
    */
    *tree = trace->t_frame[2];
    return trace->t_frame[0];
}
